// ==========================================================================
// $$
//
// @brief  Embree reader for Ward's RGBE format
// @author Jaroslav Krivanek 
// @date   Novemmber 2011

#include <iostream>
#include <cstring>
#include <cstdio>

#include "image/image.h"

extern "C" {
#include "rgbe.h"
};

namespace embree
{
	/*! read PFM file from disk */
	Ref<Image> loadRGBE(const FileName& fileName)
	{
		FILE *infp;

		if (!(infp = fopen(fileName.c_str(), "rb")))
			throw std::runtime_error("Error reading " + fileName.str());

		// Read RGBE header
		int width, height;

		if( RGBE_ReadHeader(infp, &width, &height, NULL) == RGBE_RETURN_FAILURE)
		{ 
			fclose(infp); 
			throw std::runtime_error("Could not read RGBE header from " + fileName.str());		
		}

		// Read RGBE data
		std::vector<float> data(width * height * 3);

		if( RGBE_ReadPixels_RLE( infp, &data[0], width, height) == RGBE_RETURN_FAILURE)
		{
			fclose(infp);
			throw std::runtime_error("Could not read image data from " + fileName.str());				
		}

		/*
		if( RGBE_ReadPixels(infp, &data[0], width * height) == RGBE_RETURN_FAILURE)
		{ 
			fclose(infp);
			throw std::runtime_error("Could not read image data from " + fileName.str());				
		}
		*/

		/* create image and fill with data */
		Ref<Image> img = new Image3f(width, height, fileName);

		float *dataPtr = &data[0];
		for (index_t y=0; y<height; y++)
			for (index_t x=0; x<width; x++, dataPtr+=3 )
				img->set(x,y, Col3f(dataPtr[0],dataPtr[1],dataPtr[2]) );

		fclose(infp);
		return img;
	}


	/*! store an RGBE file to disk */
	void storeRGBE(const Ref<Image>& img, const FileName& filename)
	{
		int width = (int)img->width, height = (int)img->height;
		
		FILE *fout = fopen(filename.c_str(), "wb");

		if(fout == NULL)
		{
			throw std::runtime_error("Could not open `" + filename.str() + "' for writing.");
		}

		if( RGBE_WriteHeader(fout, width, height, NULL)  != RGBE_RETURN_SUCCESS )
		{
			throw std::runtime_error("Could write RGBE header to `" + filename.str() + "'.");
		}

		// convert image data
		std::vector<float> data( width * height * 3 );

		float *dataPtr = &data[0];
		for (size_t y=0; y<img->height; y++) {
			for (size_t x=0; x<img->width; x++, dataPtr+=3) {
				Col3f c = img->get(x,y);
				dataPtr[0] = c.r;
				dataPtr[1] = c.g;
				dataPtr[2] = c.b;
			}
		}

		// write image data
		if( RGBE_WritePixels(fout, &data[0], width*height) != RGBE_RETURN_SUCCESS )
		{
			throw std::runtime_error("Could write image data to `" + filename.str() + "'.");
		}
	}

};